#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Dump the master key of a dm-crypt device that uses aes-xts-plain64."""

import argparse
import os
from pathlib import Path
import sys

from drgn import cast, container_of
from drgn.helpers.linux.block import disk_name, for_each_disk


def crypto_skcipher_alg(tfm):
    return container_of(tfm.base.__crt_alg, "struct skcipher_alg", "base")


def crypto_skcipher_ctx(tfm):
    return cast("void *", tfm.base.__crt_ctx)


def crypto_lskcipher_ctx(tfm):
    return cast("void *", tfm.base.__crt_ctx)


def aes_xts_ctx(tfm):
    AESNI_ALIGN = 16
    mask = AESNI_ALIGN - 1
    ctx = cast("unsigned long", crypto_skcipher_ctx(tfm))
    return cast("struct aesni_xts_ctx *", (ctx + mask) & ~mask)


def aes_key_from_ctx(ctx):
    words = ctx.key_enc.value_()[: ctx.key_length / 4]
    return b"".join(word.to_bytes(4, "little") for word in words)


def is_function(obj, name):
    try:
        global_ = obj.prog_[name]
    except KeyError:
        return False
    return obj == global_


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("name")
    args = parser.parse_args()

    if "/" in args.name:
        device_path = Path(args.name)
    else:
        device_path = Path("/dev/mapper") / args.name
    name = os.fsencode(device_path.resolve().name)

    for disk in for_each_disk():
        if disk_name(disk) == name:
            break
    else:
        sys.exit("target not found")

    md = cast("struct mapped_device *", disk.private_data)
    map = cast("struct dm_table *", md.map)

    if map.num_targets != 1:
        sys.exit("dm table has multiple targets")
    ti = map.targets

    if not is_function(ti.type.map, "crypt_map"):
        sys.exit("target is not dm-crypt")
    cc = cast("struct crypt_config *", ti.private)

    if cc.cipher_string.string_() != b"aes-xts-plain64":
        sys.exit("cipher is not aes-xts-plain64")

    tfm = cc.cipher_tfm.tfms[0]
    exit = crypto_skcipher_alg(tfm).exit.read_()
    if is_function(exit, "simd_skcipher_exit"):
        cryptd_tfm = cast(
            "struct simd_skcipher_ctx *", crypto_skcipher_ctx(tfm)
        ).cryptd_tfm
        cryptd_ctx = cast(
            "struct cryptd_skcipher_ctx *", crypto_skcipher_ctx(cryptd_tfm.base)
        )
        child_tfm = cryptd_ctx.child
        xts_ctx = aes_xts_ctx(cryptd_ctx.child)
        try:
            crypt_aes_ctx = xts_ctx.crypt_ctx
            tweak_aes_ctx = xts_ctx.tweak_ctx
        except AttributeError:
            # Before Linux kernel commit d148736ff17d ("crypto: x86/aesni -
            # Correct the data type in struct aesni_xts_ctx") (in v6.7), the
            # AES contexts were arrays that we need to cast.
            crypt_aes_ctx = cast("struct crypto_aes_ctx *", xts_ctx.raw_crypt_ctx)
            tweak_aes_ctx = cast("struct crypto_aes_ctx *", xts_ctx.raw_tweak_ctx)
    elif is_function(exit, "xts_exit_tfm"):
        xts_ctx = cast("struct xts_tfm_ctx *", crypto_skcipher_ctx(tfm))
        lskcipher_tfm = cast(
            "struct crypto_lskcipher **", crypto_skcipher_ctx(xts_ctx.child)
        )[0]
        cipher_tfm = cast(
            "struct crypto_cipher **", crypto_lskcipher_ctx(lskcipher_tfm)
        )[0]
        crypt_aes_ctx = cast("struct crypto_aes_ctx *", cipher_tfm.base.__crt_ctx)
        tweak_aes_ctx = cast("struct crypto_aes_ctx *", xts_ctx.tweak.base.__crt_ctx)
    else:
        sys.exit("unknown skcipher")
    print(aes_key_from_ctx(crypt_aes_ctx).hex())
    print(aes_key_from_ctx(tweak_aes_ctx).hex())


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Virtual Memory Statistics
-------------------------

The ``drgn.helpers.linux.vmstat`` module provides helpers for reading virtual
memory statistics.
"""

from drgn import IntegerLike, Object, ObjectNotFoundError, Program
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.cpumask import for_each_online_cpu
from drgn.helpers.linux.percpu import per_cpu

__all__ = (
    "global_node_page_state",
    "global_numa_event_state",
    "global_vm_event_state",
    "global_zone_page_state",
    "nr_free_pages",
    "zone_page_state",
)


@takes_program_or_default
def global_node_page_state(prog: Program, item: IntegerLike) -> int:
    """
    Get the global value of a node VM statistic.

    >>> global_node_page_state(prog["NR_FILE_PAGES"])
    2257904

    :param item: ``enum node_stat_item``
    """
    return max(prog["vm_node_stat"][item].counter.value_(), 0)


def zone_page_state(zone: Object, item: IntegerLike) -> int:
    """
    Get the value of a zone VM statistic in a single zone.

    :param zone: ``struct zone *``
    :param item: ``enum zone_stat_item``
    """
    return max(zone.vm_stat[item].counter.value_(), 0)


@takes_program_or_default
def global_zone_page_state(prog: Program, item: IntegerLike) -> int:
    """
    Get the global value of a zone VM statistic.

    >>> global_zone_page_state(prog["NR_MLOCK"])
    1562

    :param item: ``enum zone_stat_item``
    """
    return max(prog["vm_zone_stat"][item].counter.value_(), 0)


@takes_program_or_default
def nr_free_pages(prog: Program) -> int:
    """Get the number of free memory pages."""
    return global_zone_page_state(prog["NR_FREE_PAGES"])


@takes_program_or_default
def global_numa_event_state(prog: Program, item: IntegerLike) -> int:
    """
    Get the global value of a NUMA statistic.

    .. note::

        This is only valid if the kernel was compiled with ``CONFIG_NUMA``.
        Before Linux 4.14, these items (e.g., ``NUMA_HIT``) were zone
        statistics, so on kernel versions before 4.14, this is an alias of
        :func:`global_zone_page_state()`.

    :param item: ``enum numa_stat_item`` (or ``enum zone_stat_item`` on Linux <
        4.14)
    """
    # The variable was renamed from "vm_numa_stat" to "vm_numa_event" in Linux
    # kernel commit f19298b9516c ("mm/vmstat: convert NUMA statistics to basic
    # NUMA counters") (in v5.14). Additionally, before Linux kernel commit
    # 3a321d2a3dde ("mm: change the call sites of numa statistics items") (in
    # v4.14), the items were zone statistics and the variable didn't exist.
    try:
        array = prog["vm_numa_event"]
    except ObjectNotFoundError:
        try:
            array = prog["vm_numa_stat"]
        except ObjectNotFoundError:
            return global_zone_page_state(prog, item)
    return array[item].counter.value_()


@takes_program_or_default
def global_vm_event_state(prog: Program, item: IntegerLike) -> int:
    """
    Get the global value of a virtual memory event statistic.

    .. note::

        This is only valid if the kernel was compiled with
        ``CONFIG_VM_EVENT_COUNTERS``.

    :param item: ``enum vm_event_item``
    """
    event = prog["vm_event_states"].event[item]
    return sum(per_cpu(event, cpu).value_() for cpu in for_each_online_cpu(prog))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
# Generated by scripts/gen_elf_py.py.

import enum


class ET(enum.IntEnum):
    NONE = 0x0
    REL = 0x1
    EXEC = 0x2
    DYN = 0x3
    CORE = 0x4
    NUM = 0x5
    LOOS = 0xFE00
    HIOS = 0xFEFF
    LOPROC = 0xFF00
    HIPROC = 0xFFFF


class PT(enum.IntEnum):
    NULL = 0x0
    LOAD = 0x1
    DYNAMIC = 0x2
    INTERP = 0x3
    NOTE = 0x4
    SHLIB = 0x5
    PHDR = 0x6
    TLS = 0x7
    NUM = 0x8
    LOOS = 0x60000000
    GNU_EH_FRAME = 0x6474E550
    GNU_STACK = 0x6474E551
    GNU_RELRO = 0x6474E552
    GNU_PROPERTY = 0x6474E553
    GNU_SFRAME = 0x6474E554
    LOSUNW = 0x6FFFFFFA
    SUNWBSS = 0x6FFFFFFA
    SUNWSTACK = 0x6FFFFFFB
    HISUNW = 0x6FFFFFFF
    HIOS = 0x6FFFFFFF
    LOPROC = 0x70000000
    HIPROC = 0x7FFFFFFF
    MIPS_REGINFO = 0x70000000
    MIPS_RTPROC = 0x70000001
    MIPS_OPTIONS = 0x70000002
    MIPS_ABIFLAGS = 0x70000003
    PARISC_ARCHEXT = 0x70000000
    PARISC_UNWIND = 0x70000001


class SHF(enum.IntFlag):
    WRITE = 0x1
    ALLOC = 0x2
    EXECINSTR = 0x4
    MERGE = 0x10
    STRINGS = 0x20
    INFO_LINK = 0x40
    LINK_ORDER = 0x80
    OS_NONCONFORMING = 0x100
    GROUP = 0x200
    TLS = 0x400
    COMPRESSED = 0x800
    MASKOS = 0xFF00000
    MASKPROC = 0xF0000000
    GNU_RETAIN = 0x200000
    ORDERED = 0x40000000
    EXCLUDE = 0x80000000
    MIPS_GPREL = 0x10000000
    MIPS_MERGE = 0x20000000
    MIPS_ADDR = 0x40000000
    MIPS_STRINGS = 0x80000000
    MIPS_NOSTRIP = 0x8000000
    MIPS_LOCAL = 0x4000000
    MIPS_NAMES = 0x2000000
    MIPS_NODUPE = 0x1000000
    PARISC_SHORT = 0x20000000
    PARISC_HUGE = 0x40000000
    PARISC_SBP = 0x80000000
    ALPHA_GPREL = 0x10000000
    ARM_ENTRYSECT = 0x10000000
    ARM_COMDEF = 0x80000000
    IA_64_SHORT = 0x10000000
    IA_64_NORECOV = 0x20000000


class SHN(enum.IntEnum):
    UNDEF = 0x0
    LORESERVE = 0xFF00
    LOPROC = 0xFF00
    BEFORE = 0xFF00
    AFTER = 0xFF01
    HIPROC = 0xFF1F
    LOOS = 0xFF20
    HIOS = 0xFF3F
    ABS = 0xFFF1
    COMMON = 0xFFF2
    XINDEX = 0xFFFF
    HIRESERVE = 0xFFFF
    MIPS_ACOMMON = 0xFF00
    MIPS_TEXT = 0xFF01
    MIPS_DATA = 0xFF02
    MIPS_SCOMMON = 0xFF03
    MIPS_SUNDEFINED = 0xFF04
    PARISC_ANSI_COMMON = 0xFF00
    PARISC_HUGE_COMMON = 0xFF01


class SHT(enum.IntEnum):
    NULL = 0x0
    PROGBITS = 0x1
    SYMTAB = 0x2
    STRTAB = 0x3
    RELA = 0x4
    HASH = 0x5
    DYNAMIC = 0x6
    NOTE = 0x7
    NOBITS = 0x8
    REL = 0x9
    SHLIB = 0xA
    DYNSYM = 0xB
    INIT_ARRAY = 0xE
    FINI_ARRAY = 0xF
    PREINIT_ARRAY = 0x10
    GROUP = 0x11
    SYMTAB_SHNDX = 0x12
    RELR = 0x13
    NUM = 0x14
    LOOS = 0x60000000
    GNU_ATTRIBUTES = 0x6FFFFFF5
    GNU_HASH = 0x6FFFFFF6
    GNU_LIBLIST = 0x6FFFFFF7
    CHECKSUM = 0x6FFFFFF8
    LOSUNW = 0x6FFFFFFA
    SUNW_move = 0x6FFFFFFA
    SUNW_COMDAT = 0x6FFFFFFB
    SUNW_syminfo = 0x6FFFFFFC
    GNU_verdef = 0x6FFFFFFD
    GNU_verneed = 0x6FFFFFFE
    GNU_versym = 0x6FFFFFFF
    HISUNW = 0x6FFFFFFF
    HIOS = 0x6FFFFFFF
    LOPROC = 0x70000000
    HIPROC = 0x7FFFFFFF
    LOUSER = 0x80000000
    HIUSER = 0x8FFFFFFF
    MIPS_LIBLIST = 0x70000000
    MIPS_MSYM = 0x70000001
    MIPS_CONFLICT = 0x70000002
    MIPS_GPTAB = 0x70000003
    MIPS_UCODE = 0x70000004
    MIPS_DEBUG = 0x70000005
    MIPS_REGINFO = 0x70000006
    MIPS_PACKAGE = 0x70000007
    MIPS_PACKSYM = 0x70000008
    MIPS_RELD = 0x70000009
    MIPS_IFACE = 0x7000000B
    MIPS_CONTENT = 0x7000000C
    MIPS_OPTIONS = 0x7000000D
    MIPS_SHDR = 0x70000010
    MIPS_FDESC = 0x70000011
    MIPS_EXTSYM = 0x70000012
    MIPS_DENSE = 0x70000013
    MIPS_PDESC = 0x70000014
    MIPS_LOCSYM = 0x70000015
    MIPS_AUXSYM = 0x70000016
    MIPS_OPTSYM = 0x70000017
    MIPS_LOCSTR = 0x70000018
    MIPS_LINE = 0x70000019
    MIPS_RFDESC = 0x7000001A
    MIPS_DELTASYM = 0x7000001B
    MIPS_DELTAINST = 0x7000001C
    MIPS_DELTACLASS = 0x7000001D
    MIPS_DWARF = 0x7000001E
    MIPS_DELTADECL = 0x7000001F
    MIPS_SYMBOL_LIB = 0x70000020
    MIPS_EVENTS = 0x70000021
    MIPS_TRANSLATE = 0x70000022
    MIPS_PIXIE = 0x70000023
    MIPS_XLATE = 0x70000024
    MIPS_XLATE_DEBUG = 0x70000025
    MIPS_WHIRL = 0x70000026
    MIPS_EH_REGION = 0x70000027
    MIPS_XLATE_OLD = 0x70000028
    MIPS_PDR_EXCEPTION = 0x70000029
    MIPS_ABIFLAGS = 0x7000002A
    MIPS_XHASH = 0x7000002B
    PARISC_EXT = 0x70000000
    PARISC_UNWIND = 0x70000001
    PARISC_DOC = 0x70000002
    ALPHA_DEBUG = 0x70000001
    ALPHA_REGINFO = 0x70000002
    X86_64_UNWIND = 0x70000001


class STB(enum.IntEnum):
    LOCAL = 0x0
    GLOBAL = 0x1
    WEAK = 0x2
    NUM = 0x3
    LOOS = 0xA
    GNU_UNIQUE = 0xA
    HIOS = 0xC
    LOPROC = 0xD
    HIPROC = 0xF
    MIPS_SPLIT_COMMON = 0xD


class STT(enum.IntEnum):
    NOTYPE = 0x0
    OBJECT = 0x1
    FUNC = 0x2
    SECTION = 0x3
    FILE = 0x4
    COMMON = 0x5
    TLS = 0x6
    NUM = 0x7
    LOOS = 0xA
    GNU_IFUNC = 0xA
    HIOS = 0xC
    LOPROC = 0xD
    HIPROC = 0xF
    SPARC_REGISTER = 0xD
    PARISC_MILLICODE = 0xD


class STV(enum.IntEnum):
    DEFAULT = 0x0
    INTERNAL = 0x1
    HIDDEN = 0x2
    PROTECTED = 0x3



# Copyright (c) 2025, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
import argparse
import operator
import string
from typing import Any, Dict, Literal, Optional

from drgn import FaultError, IntegerLike, PlatformFlags, Program
from drgn.commands import argument, drgn_argument, mutually_exclusive_group
from drgn.commands._crash.common import (
    CrashDrgnCodeBuilder,
    _resolve_addr_or_sym,
    crash_command,
    crash_get_context,
)
from drgn.helpers.common.format import escape_ascii_string
from drgn.helpers.common.memory import IdentifiedSymbol, identify_address_all
from drgn.helpers.linux.common import IdentifiedSlabObject
from drgn.helpers.linux.mm import access_process_vm


def _crash_annotate(
    prog: Program,
    addr: int,
    level: Literal[None, "symbols", "slab", "verbose"],
    cache: Dict[Any, Any],
) -> Optional[str]:
    """Returns the crash-compatible annotation for a word of memory"""
    if level is None:
        return None
    for identified in identify_address_all(prog, addr, cache=cache):
        if isinstance(identified, IdentifiedSymbol):
            fmt = str if prog.cache.get("crash_radix", 10) == 10 else hex
            symbol = identified.symbol
            return f"{symbol.name}+{fmt(addr - symbol.address)}"
        elif isinstance(identified, IdentifiedSlabObject) and level in (
            "slab",
            "verbose",
        ):
            if identified.slab_object_info.address:
                cache_name = escape_ascii_string(
                    identified.slab_object_info.slab_cache.name.string_()
                )
            else:  # SLOB
                cache_name = "unknown slab object"
            if level == "slab":
                return f"[{cache_name}]"
            else:
                # Crash does not pad the address at all, which can help with
                # output alignment on some architectures.
                return f"[{addr:x}:{cache_name}]"
    return None


def _print_memory(
    prog: Program,
    address: IntegerLike,
    mem: bytes,
    unit: Literal[1, 2, 4, 8] = 1,
    show_ascii: bool = True,
    annotate: Literal[None, "symbols", "slab", "verbose"] = None,
    format: Literal["x", "d", "u"] = "x",
    endian: Literal["little", "big", None] = None,
    cache: Optional[Dict[Any, Any]] = None,
    indent: str = "",
    address_pad: str = "",
) -> None:
    """
    Read memory, format, and print output to stdout

    This function roughly mimics output formatting of crash's ``rd`` command. It
    allows printing in units of 1, 2, 4, or 8 bytes. Each line is always
    prefixed by the memory address, followed by 16 bytes of data formatted
    according to the options.

    :param address: starting address for the memory content
    :param mem: the memory contents. The length of the memory must be a
      multiple of the unit.
    :param unit: the number of bytes per integer to format (default is 1)
    :param show_ascii: show ASCII translations of bytes at the end of each line
    :param annotate: replace memory contents with annotations where applicable.
      Following crash's behaviors: "symbols" annotates with symbol + offset.
      "slab" includes slab cache names, and "verbose" will include slab cache
      address + name. This option can only be enabled when the unit matches the
      program word size, and when displaying as hexadecimal (not decimal).
    :param format: control the integer formatting. Use code "x" for hex (the
      default), "d" for signed integer, and "u" for unsigned integer.
    :param endian: control the endianness (default is platform's endianness)
    :param cache: opaque cache passed to :func:`identify_address_all()`
    """
    word_size = prog.address_size()
    assert unit in (1, 2, 4, 8)
    assert prog.platform is not None  # already verified by address_size()
    if len(mem) % unit != 0:
        raise ValueError("memory size must be an increment of unit")
    if annotate and unit != word_size:
        raise ValueError(
            "Annotations may only be printed for units of the platform word size"
        )
    if annotate and format != "x":
        raise ValueError("Annotations may only be printed with hexadecimal format")
    address = operator.index(address)

    byteorder: Literal["little", "big"]
    if endian is not None:
        byteorder = endian
    elif prog.platform.flags & PlatformFlags.IS_LITTLE_ENDIAN:
        byteorder = "little"
    else:
        byteorder = "big"

    chars: Dict[int, str] = {}
    if show_ascii:
        chars = {
            ord(s): s
            for s in string.ascii_letters + string.digits + string.punctuation + " "
        }

    if cache is None:
        cache = {}

    # Control the field width and padding for alignmment
    signed = False
    if format == "x":
        width = unit * 2
        value_pad = "0"
    else:
        # Crash does not correctly align decimal integers in all cases. This
        # seems like a bug: let's accurately determine the max width and get
        # correct alignment.
        if format == "d":
            widest_value = -(1 << (8 * unit - 1))
            signed = True
        else:
            widest_value = (1 << (8 * unit)) - 1
            # "u" is not actually a valid format code for Python, since ints
            # encode their sign intrinsically. Now that we've detected whether
            # signed integers are intended, set the format code correctly.
            format = "d"
        width = len(str(widest_value))
        value_pad = ""

    bytes_per_line = 16
    units_per_line = bytes_per_line // unit
    for offset in range(0, len(mem), unit):
        line_index = (offset % bytes_per_line) // unit
        if line_index == 0:
            print(f"{indent}{offset + address:{address_pad}{word_size * 2}x}: ", end="")
        value = int.from_bytes(mem[offset : offset + unit], byteorder, signed=signed)
        identified = _crash_annotate(prog, value, annotate, cache)
        if identified is not None:
            print(f" {identified:{width}s}", end="")
        else:
            print(f" {value:{value_pad}{width}{format}}", end="")

        is_end = (line_index + 1 == units_per_line) or offset + unit == len(mem)
        if is_end and show_ascii:
            # In case we didn't fill the line, pad it out so the ascii encoding
            # aligns with the ones above.
            padding = (units_per_line - line_index - 1) * (1 + width)
            print(
                " " * padding
                + "   "
                + "".join(
                    chars.get(b, ".")
                    for b in mem[offset - (offset % 16) : offset + unit]
                )
            )
        elif is_end:
            print()


def _dump_printable_memory(
    prog: Program,
    address: IntegerLike,
    count: Optional[int],
    physical: bool = False,
    kernel: bool = False,
) -> None:
    """Write ASCII printable memory contents, compatible with crash rd -a."""
    printable = set(string.printable)
    # whitespace characters other than space and newline are not printed by crash
    disallowed_ws = set(string.whitespace) - set(" \n")
    address = operator.index(address)
    if count is None:
        count = 1024 * 1024  # 1 MiB is a reasonable, if arbitrary, limit

    print_newline = True
    width = prog.address_size() * 2
    line_width = 0
    task = crash_get_context(prog)
    for offset in range(count):
        if kernel:
            byte = prog.read_u8(address + offset, physical=physical)
        else:
            byte = access_process_vm(task, address + offset, 1)[0]
        # For non-ascii bytes this gives the corresponding unicode code point.
        # That's fine because none of those characters will be present in
        # string.printable, by definition
        char = chr(byte)
        if char not in printable:
            break
        if print_newline:
            print(f"{address + offset:{width}x}:  ", end="")
            print_newline = False
            line_width = width + 3

        if char not in disallowed_ws:
            print(char, end="")
            line_width += 1
        if char == "\n":
            print_newline = True
        if line_width >= 79:
            print()
            print_newline = True

    # Print a \n to flush stdout, if we didn't just do so
    if not print_newline:
        print()


@crash_command(
    description="read memory",
    long_description="""
    This command displays the contents of memory, formatted in several different
    ways.
    """,
    arguments=(
        argument(
            "start",
            metavar="address|symbol",
            type="addr_or_sym",
            help="starting hexadecimal address, or symbol of starting address",
        ),
        argument(
            "count",
            type=int,
            help="decimal number of memory locations to display (default: 1)",
            nargs="?",
        ),
        mutually_exclusive_group(
            argument(
                "-p",
                dest="physical",
                action="store_true",
                help="address argument is a physical address",
            ),
            argument(
                "-u",
                dest="user",
                action="store_true",
                help="address argument is a user virtual address; this is only "
                "required on processors or configurations where user and kernel "
                "virtual address spaces overlap, such as s390x, or 32-bit "
                "processors with 4G:4G kernel configurations. In other cases, "
                "this can still be used as a hint.",
            ),
        ),
        mutually_exclusive_group(
            argument(
                "-d",
                dest="format",
                action="store_const",
                const="d",
                default="x",
                help="display output in signed decimal format (default is hexadecimal)",
            ),
            argument(
                "-D",
                dest="format",
                action="store_const",
                const="u",
                help="display output in unsigned decimal format (default is hexadecimal)",
            ),
        ),
        argument(
            "-s",
            dest="annotate",
            action="store_const",
            const="symbols",
            help="displays output symbolically where appropriate",
        ),
        argument(
            "-S",
            dest="annotate_extra",
            action="count",
            default=0,
            help="displays output symbolically where appropriate; if the "
            "memory contents reference a slab cache object, the name of "
            "the slab cache will be displayed in brackets. If -S is entered "
            "twice, and the memory contents reference a slab cache object, "
            "both the memory contents and the name of the slab cache will "
            "be displayed in brackets",
        ),
        argument(
            "-x",
            dest="show_ascii",
            action="store_false",
            help="do not display ASCII translation at the end of each line",
        ),
        mutually_exclusive_group(
            argument(
                "-8",
                dest="unit",
                action="store_const",
                const=1,
                help="display output in 8-bit values",
            ),
            argument(
                "-16",
                dest="unit",
                action="store_const",
                const=2,
                help="display output in 16-bit values",
            ),
            argument(
                "-32",
                dest="unit",
                action="store_const",
                const=4,
                help="display output in 32-bit values (default on 32-bit machines)",
            ),
            argument(
                "-64",
                dest="unit",
                action="store_const",
                const=8,
                help="display output in 64-bit values (default on 64-bit machines)",
            ),
            argument(
                "-a",
                dest="ascii",
                action="store_true",
                help="display output in ASCII characters if it contains printable "
                "ASCII characters",
            ),
        ),
        argument(
            "-N",
            dest="endian",
            action="store_const",
            const="big",
            help="display output in network byte order",
        ),
        argument(
            "-R",
            dest="reverse",
            action="store_true",
            help="display memory in reverse order; memory will be displayed up to and "
            "including the address argument, requiring the count argument to be greater "
            "than 1 to display memory before the specified address. Conflicts with -e.",
        ),
        argument(
            "-o",
            dest="offset",
            default=0,
            type="decimal_or_hexadecimal",
            help="offset applied to the starting address",
        ),
        argument(
            "-e",
            dest="end",
            type="hexadecimal",
            help="display memory until reaching this hexadecimal address",
        ),
        argument(
            "-r",
            dest="raw",
            type=str,
            help="dumps raw data to the specified output file; the number of bytes that "
            "are copied to the file must be specified either by a count argument or by "
            "the -e option",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_rd(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.reverse and args.end:
        raise ValueError("both -R and -e cannot be specified together")
    if args.physical and args.start[0] == "sym":
        raise ValueError("physical addresses cannot be specified by symbol")
    if args.raw and args.unit not in (None, 1):
        raise ValueError("cannot specify -r and a unit size other than 1 byte")
    if args.end is not None and args.count is not None:
        raise ValueError("only one of count and -e may be specified")
    # The -S argument can be repeated, and should result in the "annotate" field
    # being updated. Argparse can't really do it, so do it here.
    if args.annotate_extra > 1:
        args.annotate = "verbose"
    elif args.annotate_extra == 1:
        args.annotate = "slab"
    # In order for crash commands to even run, we need to have a Linux kernel
    # program, which guarantees that we have a platform. But mypy doesn't know
    # that.
    assert prog.platform is not None

    # For compatibility with crash, only enable annotations for hex output
    if args.format != "x":
        args.annotate = None
    # For compatibility with crash, disable the ascii translation when
    # annotations are enabled, or for non-hexadecimal formats
    if args.annotate is not None or args.format != "x":
        args.show_ascii = False
    # The unit default is the word size, unless reading ascii data or dumping
    # directly to a file.
    if args.unit is None:
        if args.ascii or args.raw:
            args.unit = 1
        else:
            args.unit = prog.address_size()
    # Unless we're printing ascii strings, default count to be 1. Ascii string
    # mode needs to know whether count was specified, since that is used as the
    # maximum read size.
    if args.count is None and not args.ascii:
        args.count = 1

    # We're doing something a little odd here, by reading the memory even before
    # we've emitted the drgn code. This helps us emit the correct drgn code by
    # heuristically determining whether we've been given a user or kernel
    # address. It wouldn't make sense for the --drgn operation to raise errors
    # related to looking up a symbol or reading the memory, so we need to catch
    # those errors and emit corresponding code anyway (even though it would
    # fail!). Only raise the resulting errors later, once we know we're not
    # emitting code.
    is_kernel = not args.user
    exc: Optional[Exception] = None
    try:
        start = _resolve_addr_or_sym(prog, args.start) + args.offset
    except LookupError as e:
        exc = e
        start = None
    # Only continue to read memory if we didn't fail a symbol lookup above.
    if start is not None:
        if args.end is not None:
            count = (args.end - start) // args.unit
        elif args.reverse:
            count = args.count
            start -= args.unit * (args.count - 1)
        else:
            count = args.count
        memory = None
        # Try reading kernel memory first, if the user didn't hint a userspace
        # address. Note that count may be None, in ascii mode! So fall back to a
        # one-byte read in that case. It's good enough to determine whether the
        # address is user or kernel, and the ascii mode will do its own reads
        # anyway.
        if is_kernel:
            try:
                memory = prog.read(start, (count or 1) * args.unit)
            except FaultError as e:
                exc = exc or e
        # Try reading the context process address space
        if memory is None:
            try:
                task = crash_get_context(prog)
                memory = access_process_vm(
                    task,
                    start,
                    (count or 1) * args.unit,
                )
            except FaultError as e:
                exc = exc or e
            else:
                # Success? Clear exc and note that it's a user address
                is_kernel = False
                exc = None

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.append("start = ")
        code.append_addr_or_sym(args.start)
        code.append("\n")
        if args.offset != 0:
            code.append(f"start += {args.offset}\n")
        code.append(f"unit = {args.unit}\n")

        if args.end is not None:
            code.append(
                f"""\
end = {hex(args.end)}
count = (end - start) // unit
"""
            )
        elif args.reverse:
            code.append(
                f"""\
# Print memory leading up to start:
count = {args.count or 1}
start -= unit * (count - 1)
"""
            )
        else:
            code.append(f"count = {args.count or 1}\n")

        if is_kernel and args.annotate is not None:
            code.add_from_import("drgn.helpers.common.memory", "print_annotated_memory")
            code.append(
                f"""\
print_annotated_memory(start, count * unit, physical={args.physical})
"""
            )
        elif is_kernel:
            code.append(
                f"""\
memory = prog.read(start, count * unit, physical={args.physical})
"""
            )
        else:
            code.add_from_import("drgn.helpers.linux.mm", "access_process_vm")
            code.add_from_import("drgn.helpers.linux.pid", "find_task")
            code.append(
                f"""\
task = find_task(prog, {task.pid.value_()})
memory = access_process_vm(task, start, count)
"""
            )

        if args.raw:
            code.append(
                f"""\
with open({args.raw!r}, "wb") as f:
    f.write(memory)
"""
            )
        code.print()
        return
    if exc:
        raise exc
    assert memory is not None
    assert start is not None

    if args.raw:
        with open(args.raw, "wb") as f:
            f.write(memory)
        print(f"{len(memory)} bytes copied from 0x{start:x} to {args.raw}")
    elif args.ascii:
        _dump_printable_memory(
            prog, start, count, physical=args.physical, kernel=is_kernel
        )
    else:
        _print_memory(
            prog,
            start,
            memory,
            unit=args.unit,
            annotate=args.annotate,
            format=args.format,
            endian=args.endian,
            show_ascii=args.show_ascii,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Crash commands for kernel modules and debugging symbols."""

import argparse
import contextlib
import logging
from typing import Any, Iterator, List, Sequence

from drgn import Program, RelocatableModule
from drgn.commands import (
    CommandArgumentError,
    _repr_black,
    argument,
    drgn_argument,
    mutually_exclusive_group,
)
from drgn.commands._crash.common import CrashDrgnCodeBuilder, crash_command
from drgn.helpers.common.format import CellFormat, print_table
from drgn.helpers.linux.module import module_taints


@contextlib.contextmanager
def _log_debug() -> Iterator[None]:
    logger = logging.getLogger("drgn")
    old_level = logger.level
    try:
        logger.setLevel(logging.DEBUG)
        yield
    finally:
        logger.setLevel(old_level)


@crash_command(
    description="module information and loading debugging symbols",
    arguments=(
        mutually_exclusive_group(
            argument(
                "-s",
                dest="load",
                nargs="+",
                metavar=("MODULE", "FILE"),
                help="""
                load debugging symbols for the given module from the given
                file, or from the default locations if no file is given
                """,
            ),
            argument(
                "-S",
                dest="load_all",
                metavar="DIRECTORY",
                nargs="?",
                default=argparse.SUPPRESS,
                help="""
                load debugging symbols for all modules from the given
                directory, or from the default locations if no directory is
                given
                """,
            ),
            argument(
                "-t",
                dest="taints",
                action="store_true",
                help="display tainted modules",
            ),
        ),
        drgn_argument,
    ),
)
def _crash_cmd_mod(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    # argparse doesn't support nargs ranges, so we have to check this manually.
    if args.load is not None and len(args.load) > 2:
        raise CommandArgumentError("-s takes at most one file")

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)

        if hasattr(args, "load_all"):
            if args.load_all is not None:
                code.append(
                    f"""\
old_kernel_directories = prog.debug_info_options.kernel_directories
try:
    prog.debug_info_options.kernel_directories = ({_repr_black(args.load_all)},)
    """
                )
            code.append("prog.load_default_debug_info()\n")
            if args.load_all is None:
                code.append("\n")
            else:
                code.append(
                    """\
finally:
    prog.debug_info_options.kernel_directories = old_kernel_directories

"""
                )

        code.add_from_import("drgn", "RelocatableModule")
        code.append(
            """\
for module, _ in prog.loaded_modules():
    if not isinstance(module, RelocatableModule):
        # Skip vmlinux.
        continue
"""
        )
        if args.load is not None:
            code.append(
                f"""
    if module.name != {_repr_black(args.load[0])}:
        continue

"""
            )
            if len(args.load) == 1:
                code.append("    prog.load_module_debug_info(module)\n\n")
            else:
                code.append(f"    module.try_file({_repr_black(args.load[1])})\n\n")

        code.append("    struct_module = module.object\n")

        if args.taints:
            code.add_from_import("drgn.helpers.linux.module", "module_taints")
            code.append("    taints = module_taints(struct_module)\n")
        else:
            code.append(
                """\
    name = module.name
    text_base = module.address
    size = sum(end - start for start, end in module.address_ranges)
    object_file = module.debug_file_path
"""
            )
        code.print()
        return

    if args.taints:
        rows: List[Sequence[Any]] = [("NAME", "TAINTS")]
        for module, _ in prog.loaded_modules():
            if not isinstance(module, RelocatableModule):
                continue
            taints = module_taints(module.object)
            if taints:
                rows.append((module.name, taints))
        if len(rows) > 1:
            print_table(rows)
        else:
            print("no tainted modules")
        return

    if hasattr(args, "load_all"):
        old_kernel_directories = prog.debug_info_options.kernel_directories
        try:
            if args.load_all is not None:
                prog.debug_info_options.kernel_directories = (args.load_all,)
            with _log_debug():
                prog.load_default_debug_info()
        finally:
            prog.debug_info_options.kernel_directories = old_kernel_directories

    rows = [
        (
            CellFormat("MODULE", "^"),
            "NAME",
            CellFormat("TEXT_BASE", "^"),
            CellFormat("SIZE", ">"),
            "OBJECT FILE",
        ),
    ]
    for module, _ in prog.loaded_modules():
        if not isinstance(module, RelocatableModule):
            continue

        if args.load is not None:
            if args.load[0] != module.name:
                continue
            with _log_debug():
                if len(args.load) == 1:
                    prog.load_module_debug_info(module)
                else:
                    module.try_file(args.load[1])

        object_file = module.debug_file_path
        if object_file is None:
            object_file = "(not loaded)"
            if getattr(module.object, "kallsyms", None):
                object_file += "  [CONFIG_KALLSYMS]"

        if module.address_ranges is None:
            size = 0
        else:
            size = sum(end - start for start, end in module.address_ranges)

        rows.append(
            (
                CellFormat(module.object.value_(), "^x"),
                module.name,
                CellFormat(module.address, "^x"),
                size,
                object_file,
            )
        )
    print_table(rows)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Common
------

The ``drgn.helpers.common`` package provides helpers that can be used with any
program. The helpers are available from the individual modules in which they
are defined and from this top-level package. E.g., the following are both
valid:

>>> from drgn.helpers.common.memory import identify_address
>>> from drgn.helpers.common import identify_address

Some of these helpers may have additional program-specific behavior but are
otherwise generic.
"""

import importlib
import pkgutil
from typing import List

__all__: List[str] = []
for _module_info in pkgutil.iter_modules(__path__, prefix=__name__ + "."):
    _submodule = importlib.import_module(_module_info.name)
    _submodule_all = getattr(_submodule, "__all__", ())
    __all__.extend(_submodule_all)
    for _name in _submodule_all:
        globals()[_name] = getattr(_submodule, _name)



#!/usr/bin/python3
# Copyright (c) 2023, Oracle and/or its affiliates.
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
"""
Drgn CLI, but using ptpython rather than the standard code.interact()

NOTE: this is definitely a bit of a hack, using implementation details of Drgn
*and* ptpython. It may break at any time, but it is also quite useful, and this
makes it worth sharing.

Requires: "pip install ptpython" which brings in pygments and prompt_toolkit
"""
import functools
import os
import shutil
from typing import Any, Dict, Set

from prompt_toolkit.completion import Completer
from prompt_toolkit.formatted_text import PygmentsTokens
from ptpython import embed
from ptpython.repl import run_config
from pygments.lexers.c_cpp import CLexer

import drgn
import drgn.cli


class DummyForRepr:
    """
    A dummy class to pass back to _format_result_output() that pretends to have
    the given repr()
    """

    def __init__(self, s):
        self.s = s

    def __repr__(self):
        return self.s


class DummyForPtRepr:
    """A similar dummy class for the __pt_repr__() method."""

    def __init__(self, s):
        self.s = s

    def __pt_repr__(self):
        return self.s


def _maybe_c_format(s):
    """Given a string, try to use pygments to highlight it it as a C string."""
    try:
        tokens = CLexer().get_tokens_unprocessed(s)
        formatted = PygmentsTokens([(tokentype, value) for index, tokentype, value in tokens])
        to_format = DummyForPtRepr(formatted)
    except Exception as e:
        to_format = DummyForRepr(s)
    return to_format


@functools.lru_cache(maxsize=1)
def _object_fields() -> Set[str]:
    return set(dir(drgn.Object))


class ReorderDrgnObjectCompleter(Completer):
    """A completer which puts Object member fields above Object defaults"""

    def __init__(self, c: Completer):
        self.c = c

    def get_completions(self, document, complete_event):
        completions = list(self.c.get_completions(document, complete_event))
        if not completions:
            return completions
        text = completions[0].text
        member_fields = []
        # If the first completion is "absent_", it is *very likely* that we are
        # now looking at the completion of on Object.  Move the default Object
        # attributes to the end of the list so that we get the struct attributes
        if text == "absent_":
            fields = _object_fields()
            for i in reversed(range(len(completions))):
                text = completions[i].text
                if text not in fields:
                    member_fields.append(completions[i])
                    del completions[i]
            return list(reversed(member_fields)) + completions
        return completions


def configure(repl) -> None:
    """
    Muck around with the internals of PythonRepl so that we will special case the
    drgn data structures, similar to how drgn messes with sys.displayhook. We can
    do C syntax highlighting too, which is really nice.

    This also automatically runs the default config file:
    ~/.config/ptpython/config.py
    """
    _format_result_output_orig = repl._format_result_output

    def _format_result_output(result: object):
        if isinstance(result, drgn.Object):
            try:
                s = result.format_(columns=shutil.get_terminal_size((0, 0)).columns)
                to_format = _maybe_c_format(s)
            except drgn.FaultError:
                to_format = DummyForRepr(repr(result))
        elif isinstance(result, (drgn.StackFrame, drgn.StackTrace)):
            to_format = DummyForRepr(str(result))
        elif isinstance(result, drgn.Type):
            to_format = _maybe_c_format(str(result))
        else:
            to_format = result
        return _format_result_output_orig(to_format)

    repl._format_result_output = _format_result_output
    run_config(repl)
    repl._completer = ReorderDrgnObjectCompleter(repl._completer)
    repl.completer = ReorderDrgnObjectCompleter(repl.completer)


def interact(local: Dict[str, Any], banner: str):
    histfile = os.path.expanduser("~/.drgn_history.ptpython")
    print(banner)
    embed(globals=local, history_filename=histfile, title="drgn", configure=configure)


if __name__ == "__main__":
    # Muck around with the internals of drgn: swap out run_interactive() with our
    # ptpython version, and then call main as if nothing happened.
    drgn.cli.interact = interact
    drgn.cli._main()



#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Dump the list of memory regions exposed by /proc/kcore."""

from drgn import cast
from drgn.helpers.linux.list import list_for_each_entry

kcore_type = prog.type("enum kcore_type")
for entry in list_for_each_entry(
    "struct kcore_list", prog["kclist_head"].address_of_(), "list"
):
    print(
        f"{cast(kcore_type, entry.type).format_(type_name=False)} {hex(entry.addr)} {hex(entry.size)}"
    )



# (C) Copyright IBM Corp. 2026
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Sysfs
-----

The ``drgn.helpers.linux.sysfs`` module provides helpers for working with
the sysfs interface built on top of kernfs.
"""

from typing import List, Optional

from drgn import NULL, Object, Program, cast, container_of
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.kernfs import (
    _kernfs_node_type,
    kernfs_children,
    kernfs_parent,
    kernfs_path,
    kernfs_walk,
)

__all__ = (
    "sysfs_lookup_node",
    "sysfs_lookup_kobject",
    "sysfs_lookup",
    "sysfs_listdir",
    "kobject_path",
)


@takes_program_or_default
def sysfs_lookup_node(prog: Program, path: str) -> Object:
    """
    Look up a ``struct kernfs_node *`` for a given sysfs path.

    The path may be provided either relative to ``/sys`` (e.g.,
    ``devices/system/cpu``) or as an absolute path beginning with
    ``/sys``.

    If the path is empty or refers to ``/sys`` itself, this returns
    the sysfs root node.

    :param path: Sysfs path (absolute or relative to ``/sys``)
    :return: ``struct kernfs_node *`` or ``NULL``
    """
    root_kn = prog["sysfs_root_kn"]

    if path.startswith("/"):
        path = path.lstrip("/")
        if path == "sys":
            path = ""
        elif path.startswith("sys/"):
            path = path[4:].lstrip("/")
        else:
            return NULL(prog, "struct kernfs_node *")

    return kernfs_walk(root_kn, path)


@takes_program_or_default
def sysfs_lookup_kobject(prog: Program, path: str) -> Optional[Object]:
    """
    Look up the ``struct kobject *`` corresponding to a sysfs path.

    If the path refers to a directory, this returns ``kn->priv``.
    If the path refers to a file (attribute), this returns the
    parent kobject (i.e., the parent node’s ``priv``)

    :param path: Sysfs path (absolute or relative to ``/sys``)
    :return: ``struct kobject *`` or ``NULL``
    """
    kn = sysfs_lookup_node(prog, path)
    if not kn:
        return NULL(prog, "struct kobject *")

    if _kernfs_node_type(kn, "KERNFS_DIR"):
        return cast("struct kobject *", kn.priv)

    parent = kernfs_parent(kn)
    if not parent:
        return NULL(prog, "struct kobject *")

    return cast("struct kobject *", parent.priv)


@takes_program_or_default
def sysfs_lookup(prog: Program, path: str) -> Optional[Object]:
    """
    Look up the object represented by a sysfs path.

    Resolve the ``struct kobject`` corresponding to a sysfs entry and,
    when possible, return the kernel structure which contains that
    kobject.

    Note: This may return more specific types for other cases in the future.

    Depending on the kobject type, this helper may return:

    - ``struct device *``
    - ``struct class *``
    - ``struct bus_type *``
    - ``struct device_driver *``
    - ``struct module_kobject *``

    If the kobject type does not correspond to a known container
    structure, the function returns the ``struct kobject *``.

    :param path: Sysfs path (absolute or relative to ``/sys``)
    :return: Corresponding container structure pointer or
        ``struct kobject *`` or ``NULL``
    """
    kobj = sysfs_lookup_kobject(prog, path)
    if not kobj:
        return NULL(prog, "struct kobject *")

    # device
    try:
        device_ktype = prog["device_ktype"].address_of_()
    except KeyError:
        pass
    else:
        if kobj.ktype == device_ktype:
            return container_of(kobj, "struct device", "kobj")

    # module
    try:
        module_ktype = prog["module_ktype"].address_of_()
    except KeyError:
        pass
    else:
        if kobj.ktype == module_ktype:
            return container_of(kobj, "struct module_kobject", "kobj")

    # driver
    try:
        driver_ktype = prog["driver_ktype"].address_of_()
    except KeyError:
        pass
    else:
        if kobj.ktype == driver_ktype:
            drv_priv = container_of(kobj, "struct driver_private", "kobj")
            return drv_priv.driver

    # class
    try:
        class_ktype = prog["class_ktype"].address_of_()
    except KeyError:
        pass
    else:
        if kobj.ktype == class_ktype:
            kset = container_of(kobj, "struct kset", "kobj")
            subsys = container_of(kset, "struct subsys_private", "subsys")
            return getattr(subsys, "class")

    # bus
    try:
        bus_ktype = prog["bus_ktype"].address_of_()
    except KeyError:
        pass
    else:
        if kobj.ktype == bus_ktype:
            kset = container_of(kobj, "struct kset", "kobj")
            subsys = container_of(kset, "struct subsys_private", "subsys")
            return subsys.bus

    return kobj


@takes_program_or_default
def sysfs_listdir(prog: Program, path: str) -> List[bytes]:
    """
    List the children of a sysfs directory.

    :param path: Sysfs directory path (absolute or relative to ``/sys``)
    :return: List of child entry names
    :raises ValueError: If path is not found or not a directory
    """
    kn = sysfs_lookup_node(prog, path)
    if not kn:
        raise ValueError(f"{path}: not found")

    return [child.name.string_() for child in kernfs_children(kn)]


def kobject_path(kobj: Object) -> bytes:
    """
    Return the sysfs path corresponding to a ``struct kobject *``.

    :param kobj: ``struct kobject *``
    :return: Absolute path in sysfs
    """
    kn = kobj.sd
    path = kernfs_path(kn)

    if path == b"/":
        return b"/sys"

    return b"/sys" + path



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Nulls Lists
-----------

The ``drgn.helpers.linux.list_nulls`` module provides helpers for working with
the special version of lists (``struct hlist_nulls_head`` and ``struct
hlist_nulls_node``) in :linux:`include/linux/list_nulls.h` where the end of
list is not a ``NULL`` pointer, but a "nulls" marker.
"""

from typing import Iterator, Union

from drgn import Object, Type, container_of

__all__ = (
    "hlist_nulls_empty",
    "hlist_nulls_for_each_entry",
    "is_a_nulls",
)


def is_a_nulls(pos: Object) -> bool:
    """
    Return whether a a pointer is a nulls marker.

    :param pos: ``struct hlist_nulls_node *``
    """
    return bool(pos.value_() & 1)


def hlist_nulls_empty(head: Object) -> bool:
    """
    Return whether a nulls hash list is empty.

    :param head: ``struct hlist_nulls_head *``
    """
    return is_a_nulls(head.first)


def hlist_nulls_for_each_entry(
    type: Union[str, Type], head: Object, member: str
) -> Iterator[Object]:
    """
    Iterate over all the entries in a nulls hash list.

    :param type: Entry type.
    :param head: ``struct hlist_nulls_head *``
    :param member: Name of list node member in entry type.
    :return: Iterator of ``type *`` objects.
    """
    type = head.prog_.type(type)
    pos = head.first
    while not is_a_nulls(pos):
        yield container_of(pos, type, member)
        pos = pos.next



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Radix Trees
-----------

The ``drgn.helpers.linux.radixtree`` module provides helpers for working with
radix trees from :linux:`include/linux/radix-tree.h`.

.. seealso::

    `XArrays`_, which were introduced in Linux 4.20 as a replacement for radix
    trees.
"""

from typing import Iterator, Tuple

from drgn import IntegerLike, Object
from drgn.helpers.linux.xarray import xa_for_each, xa_load

__all__ = (
    "radix_tree_for_each",
    "radix_tree_lookup",
)


def radix_tree_lookup(root: Object, index: IntegerLike) -> Object:
    """
    Look up the entry at a given index in a radix tree.

    :param root: ``struct radix_tree_root *``
    :param index: Entry index.
    :return: ``void *`` found entry, or ``NULL`` if not found.
    """
    return xa_load(root, index)


def radix_tree_for_each(root: Object) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over all of the entries in a radix tree.

    :param root: ``struct radix_tree_root *``
    :return: Iterator of (index, ``void *``) tuples.
    """
    return xa_for_each(root)



#!/usr/bin/env drgn

# Copyright (c) 2025 NVIDIA Corporation & Affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later

help="""
Detect leaked page_pool pages by scanning through all the pages.

Has options for peeking into the page memory and showing the struct page.
"""


import argparse
from drgn import FaultError
from drgn.helpers.common.memory import (
        print_annotated_memory
)
from drgn.helpers.linux.mm import (
    for_each_page,
    page_to_virt
)
from drgn.helpers.linux.net import is_pp_page


def get_opts():
    parser = argparse.ArgumentParser(description=help)
    parser.add_argument(
        "-l", "--peek", default=100, type=int, help="Peek into page given amount of bytes.")
    parser.add_argument(
        "-s", "--show", default=False, action="store_true", help="Show page struct.")

    args = parser.parse_args()
    return args


opt = get_opts()

for page in for_each_page():
        try:
            if is_pp_page(page) and page.pp.user.detach_time > 0:
                if opt.show:
                    print(page)
                else:
                    print(f"Leaked page: {hex(page)}")
                if opt.peek > 0:
                    print("Page content: ")
                    print_annotated_memory(page_to_virt(page), opt.peek)
        except FaultError:
            continue




# Copyright (c) 2024, Oracle and/or its affiliates.
"""
negdentdelete - remove negative dentries from a directory

Normally, there aren't many good ways to get rid of negative dentries. You
could:

1. Delete the entire directory containing them. This works, but if there are
   real files in it, then it's a lot of work.
2. Use drop_caches. This option is unfortunately global and so you can't target
   specific files.
3. Wait until your memory fills up and the LRU starts to handle it...

This script can help you target a specific directory and remove all negative
dentries within it. Note that while the script is reasonably safe, it can't be
100% reliable: iterating over a frequently changing linked list in the kernel
may go awry. The script also creates and unlinks files within the target
directory, so the user must have permission to do so. If the script runs as
root, then you should ensure that root-owned files inside the directory won't
cause any problems.
"""
import argparse
import os
import time
from typing import Iterator, List

from drgn import Object, Program
from drgn.helpers.linux.fs import path_lookup
from drgn.helpers.linux.list import hlist_for_each_entry, list_for_each_entry


def for_each_child_dentry(dentry: Object) -> Iterator[Object]:
    try:
        # da549bdd15c29 ("dentry: switch the lists of children to hlist")
        return hlist_for_each_entry(
            "struct dentry",
            dentry.d_children.address_of_(),
            "d_sib",
        )
    except AttributeError:
        return list_for_each_entry(
            "struct dentry", dentry.d_subdirs.address_of_(), "d_child"
        )


def yield_negative_dentries(
    prog: Program, dir_: str, chunk_size: int = 10000
) -> Iterator[List[bytes]]:
    parent = path_lookup(prog, dir_).dentry
    negdent_names = []

    for child in for_each_child_dentry(parent):
        # Do this at the top of the loop so that there's less of a chance of the
        # current child being freed out from under us. It's sort of like
        # list_for_each_entry_safe(). Of course, there's no guarantee that this
        # dentry will still be here when we come back (we're not holding a
        # reference, after all). But at least we're not actively freeing the
        # dentry we're currently looking at.
        if len(negdent_names) >= chunk_size:
            yield negdent_names
            negdent_names = []

        if not child.d_inode:
            negdent_names.append(child.d_name.name.string_())
    if negdent_names:
        yield negdent_names


def remove_negative_dentries(
    dir_: str, names: List[bytes], verbose: bool = False
) -> None:
    dir_fd = os.open(dir_, os.O_PATH)
    try:
        for name in names:
            # When a file is open and it is unlinked, its associated dentry cannot
            # remain a part of the dentry cache (since a new file of the same name
            # could be created). So, it is removed from the dentry hash table so it
            # can no longer be looked up (see d_delete() in fs/dcache.c).
            #
            # When the file is closed, dput() will find that the dentry is
            # unhashed, and so it will immediately free it. Thus, creating a file
            # with the same name as a negative dentry, unlinking, and then closing
            # it, is a sneaky way of removing the cached negative dentry. While
            # this isn't ideal (creating the file does result in some I/O), it is
            # still remarkably quick.
            fd = os.open(name, os.O_RDONLY | os.O_CREAT, dir_fd=dir_fd)
            os.unlink(name, dir_fd=dir_fd)
            os.close(fd)
            if verbose:
                print(name.decode(), fd)
    finally:
        os.close(dir_fd)


def main(prog: Program):
    parser = argparse.ArgumentParser(
        description="remove negative dentries from a directory"
    )
    parser.add_argument(
        "directory",
        help="directory to clear negative dentries from",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="print each dentry we delete (much slower!)",
    )
    parser.add_argument(
        "--chunk-size",
        type=int,
        default=10000,
        help="number of negative dentries to read at a time",
    )
    args = parser.parse_args()
    directory = os.path.abspath(args.directory)
    removed = 0
    start = time.time()
    for batch in yield_negative_dentries(prog, directory, args.chunk_size):
        remove_negative_dentries(directory, batch, verbose=args.verbose)
        removed += len(batch)
    total = time.time() - start
    dps = removed / total
    print(f"removed {removed} negative dentries in {total:.2f}s ({dps:.2f}/s)")


if __name__ == "__main__":
    prog: Program
    main(prog)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

# Filesystem-related commands.

import argparse
import operator
import os
import sys
from typing import Any, List

from drgn import Object, Program
from drgn.commands import _repr_black, argument, drgn_argument, mutually_exclusive_group
from drgn.commands._crash._kmem import _print_pages_default_members
from drgn.commands._crash.common import (
    CrashDrgnCodeBuilder,
    _crash_foreach_subcommand,
    _TaskSelector,
    crash_command,
    crash_get_context,
    print_task_header,
)
from drgn.helpers.common.format import CellFormat, escape_ascii_string, print_table
from drgn.helpers.linux.fs import (
    address_space_for_each_page,
    d_path,
    decode_file_type,
    for_each_file,
    for_each_mount,
    mount_dst,
    mount_fstype,
    mount_src,
)


def _print_dentry(prog: Program, address: int, drgn_arg: bool) -> None:
    if drgn_arg:
        sys.stdout.write(
            f"""\
from drgn import Object
from drgn.helpers.linux.fs import d_path, decode_file_type


dentry = Object(prog, "struct dentry *", {hex(address)})
inode = dentry.d_inode
sb = inode.i_sb
type = decode_file_type(inode.i_mode)
path = d_path(dentry)
"""
        )
        return

    dentry = Object(prog, "struct dentry *", address)
    inode = dentry.d_inode.read_()
    sb = inode.i_sb.read_()
    print_table(
        (
            (
                CellFormat("DENTRY", "^"),
                CellFormat("INODE", "^"),
                CellFormat("SUPERBLK", "^"),
                "TYPE",
                "PATH",
            ),
            (
                CellFormat(dentry.value_(), "^x"),
                CellFormat(inode.value_(), "^x"),
                CellFormat(sb.value_(), "^x"),
                decode_file_type(inode.i_mode),
                escape_ascii_string(d_path(dentry), escape_backslash=True),
            ),
        )
    )


def _print_inode(prog: Program, address: int, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn", "Object")
        code.add_from_import("drgn.helpers.linux.fs", "inode_for_each_page")
        code.add_from_import(
            "drgn.helpers.linux.mm",
            "decode_page_flags_value",
            "page_flags",
            "page_to_phys",
        )
        code.append(
            f"""\
inode = Object(prog, "struct inode *", {hex(address)})
i_mapping = inode.i_mapping
nrpages = i_mapping.nrpages

for index, page in inode_for_each_page(inode):
    physical = page_to_phys(page)
    cnt = page._refcount.counter
    flags = page_flags(page)
    decoded_flags = decode_page_flags_value(flags)
"""
        )
        return code.print()

    inode = Object(prog, "struct inode *", address)
    mapping = inode.i_mapping.read_()
    print_table(
        (
            (CellFormat("INODE", "^"), CellFormat("NRPAGES", ">")),
            (CellFormat(inode.value_(), "^x"), mapping.nrpages.value_()),
        )
    )

    print()
    mapping_value = mapping.value_()
    _print_pages_default_members(
        prog,
        address_space_for_each_page(mapping),
        get_page=operator.itemgetter(1),
        get_mapping=lambda _: mapping_value,
        get_index=operator.itemgetter(0),
    )


@_crash_foreach_subcommand(
    arguments=(
        argument(
            "-c",
            dest="cache",
            action="store_true",
        ),
        argument(
            "-R",
            dest="reference",
        ),
        drgn_argument,
    ),
)
def _crash_foreach_files(
    task_selector: _TaskSelector, args: argparse.Namespace
) -> None:
    prog = task_selector.prog

    reference_int = None
    if args.reference is not None:
        try:
            reference_int = int(args.reference, 10)
            reference_int_base = 10
        except ValueError:
            try:
                reference_int = int(args.reference, 16)
                reference_int_base = 16
            except ValueError:
                pass

        reference_path = os.fsencode(args.reference)

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)

        if reference_int is not None:
            if reference_int_base == 10:
                code.append(f"reference = {reference_int}\n")
            else:
                code.append(f"reference = {hex(reference_int)}\n")
        if args.reference is not None:
            code.add_import("os")
            code.append(
                f"reference_path = os.fsencode({_repr_black(args.reference)})\n\n"
            )

        with task_selector.begin_task_loop(code):
            code.append_task_header()

            code.add_from_import(
                "drgn.helpers.linux.fs", "d_path", "decode_file_type", "for_each_file"
            )

            code.append(
                """
root = d_path(task.fs.root)
cwd = d_path(task.fs.pwd)

"""
            )
            if args.reference is not None:
                code.append(
                    """\
is_match = (reference_path in root) or (reference_path in cwd)

"""
                )

            code.append("for fd, file in for_each_file(task):\n")
            if args.cache:
                code.append(
                    """\
    inode = file.f_inode
    i_mapping = inode.i_mapping
    nrpages = i_mapping.nrpages
"""
                )
            else:
                code.append(
                    """\
    dentry = file.f_path.dentry
    inode = file.f_inode
"""
                )
            code.append(
                """\
    type = decode_file_type(inode.i_mode)
    path = d_path(file.f_path)
"""
            )
            if reference_int is not None:
                code.append(
                    """\
    is_match = (
        fd == reference
"""
                )
                if args.cache:
                    code.append(
                        """\
        or inode.value_() == reference
        or i_mapping.value_() == reference
"""
                    )
                else:
                    code.append(
                        """\
        or file.value_() == reference
        or dentry.value_() == reference
        or inode.value_() == reference
"""
                    )
                code.append(
                    """\
        or reference_path in path
    )
"""
                )
            elif args.reference is not None:
                code.append(
                    """\
    is_match = reference_path in path
"""
                )
        return code.print()

    first = True
    for task in task_selector.tasks():
        row: List[Any] = [CellFormat("FD", ">")]
        if args.cache:
            row.append(CellFormat("INODE", "^"))
            row.append(CellFormat("I_MAPPING", "^"))
            row.append(CellFormat("NRPAGES", ">"))
        else:
            row.append(CellFormat("FILE", "^"))
            row.append(CellFormat("DENTRY", "^"))
            row.append(CellFormat("INODE", "^"))
        row.append("TYPE")
        row.append("PATH")
        rows = [row]

        for fd, file in for_each_file(task):
            f_path = file.f_path.read_()
            inode = file.f_inode.read_()
            if args.cache:
                mapping = inode.i_mapping.read_()
            else:
                dentry = f_path.dentry.read_()
            path = d_path(f_path)

            if args.reference is not None and not (
                (
                    reference_int is not None
                    and (
                        fd == reference_int
                        or inode.value_() == reference_int
                        # Crash only matches against the columns that it prints
                        # depending on -c.
                        or (args.cache and mapping.value_() == reference_int)
                        or (not args.cache and file.value_() == reference_int)
                        or (not args.cache and dentry.value_() == reference_int)
                    )
                )
                or reference_path in path
            ):
                continue

            row = [fd]
            if args.cache:
                row.append(CellFormat(inode.value_(), "^x"))
                row.append(CellFormat(mapping.value_(), "^x"))
                row.append(mapping.nrpages.value_())
            else:
                row.append(CellFormat(file.value_(), "^x"))
                row.append(CellFormat(dentry.value_(), "^x"))
                row.append(CellFormat(inode.value_(), "^x"))
            row.append(decode_file_type(inode.i_mode))
            row.append(escape_ascii_string(path, escape_backslash=True))
            rows.append(row)

        fs = task.fs
        root_path = d_path(fs.root)
        cwd_path = d_path(fs.pwd)
        if (
            args.reference is None
            or len(rows) > 1
            or reference_path in root_path
            or reference_path in cwd_path
        ):
            if first:
                first = False
            else:
                print()
            print_task_header(task)

            fs = task.fs
            print(
                f"ROOT: {escape_ascii_string(root_path, escape_backslash=True)}"
                f"  CWD: {escape_ascii_string(cwd_path, escape_backslash=True)}"
            )

            if len(rows) > 1:
                print_table(rows)
            elif args.reference is None:
                print("No open files")


@crash_command(
    description="file information",
    long_description="""
    Show file information. By default, display the root directory, current
    working directory, and all open files of one or more tasks, including the
    file descriptor, file structure, dentry, inode, file type, and path.
    """,
    arguments=(
        mutually_exclusive_group(
            argument(
                "-d",
                dest="dentry",
                type="hexadecimal",
                help="""
                display the inode, super block, file type, and full path of a
                dentry, given as a hexadecimal address
                """,
            ),
            argument(
                "-p",
                dest="inode",
                type="hexadecimal",
                help="""
                display the pages in the page cache of an inode, given as a
                hexadecimal address
                """,
            ),
            argument(
                "tasks",
                metavar="pid|task",
                type="pid_or_task",
                nargs="*",
                # Work around https://github.com/python/cpython/issues/72795
                # before Python 3.13.
                default=[],
                help="""
                display open files for this task, given as either a decimal process
                ID or a hexadecimal ``task_struct`` address. May be given multiple
                times. Defaults to the current context
                """,
            ),
        ),
        argument(
            "-c",
            dest="cache",
            action="store_true",
            help="""
            for each open file, display the i_mapping and number of cached
            pages instead of the file structure and dentry
            """,
        ),
        argument(
            "-R",
            dest="reference",
            help="""
            search for references to this file descriptor, path, or dentry,
            inode, address_space, or file structure address
            """,
        ),
        drgn_argument,
    ),
)
def _crash_cmd_files(
    prog: Program,
    name: str,
    args: argparse.Namespace,
    *,
    parser: argparse.ArgumentParser,
    **kwargs: Any,
) -> None:
    # -c and -R are mutually exclusive with -d and -p, but not with each other
    # or with tasks. argparse can't express this.
    if (args.cache or args.reference is not None) and (args.dentry or args.inode):
        parser.error("-c/-R not allowed with -d/-p")

    if args.dentry is not None:
        return _print_dentry(prog, args.dentry, args.drgn)
    elif args.inode is not None:
        return _print_inode(prog, args.inode, args.drgn)

    if not args.tasks:
        args.tasks.append(None)
    return _crash_foreach_files(_TaskSelector(prog, args.tasks), args)


@crash_command(
    description="mounted filesystems",
    long_description="List mounted filesystems.",
    arguments=(
        argument(
            "-n",
            dest="task",
            metavar="pid|task",
            type="pid_or_task",
            help="list mounted filesystems in the namespace of a task, "
            "given as either a decimal process ID "
            "or a hexadecimal ``task_struct`` address. "
            "Defaults to the mount namespace of the current context",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_mount(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import(
            "drgn.helpers.linux.fs",
            "for_each_mount",
            "mount_dst",
            "mount_fstype",
            "mount_src",
        )
        # Avoid the context/mount namespace noise if -n wasn't given and the
        # current context is in the initial mount namespace.
        if (
            args.task is None
            and crash_get_context(prog).nsproxy.mnt_ns
            == prog["init_task"].nsproxy.mnt_ns
        ):
            code.append("for mnt in for_each_mount():\n")
        else:
            code.append_crash_context(args.task)
            code.append(
                """\

mnt_ns = task.nsproxy.mnt_ns
for mnt in for_each_mount(mnt_ns):
"""
            )

        code.append(
            """\
    superblock = mnt.mnt.mnt_sb
    fstype = mount_fstype(mnt)
    devname = mount_src(mnt)
    dirname = mount_dst(mnt)
"""
        )
        code.print()
        return

    task = crash_get_context(prog, args.task)
    rows = [
        (
            CellFormat("MOUNT", "^"),
            CellFormat("SUPERBLK", "^"),
            "TYPE",
            "DEVNAME",
            "DIRNAME",
        )
    ]
    for mnt in for_each_mount(task.nsproxy.mnt_ns):
        rows.append(
            (
                CellFormat(mnt.value_(), "^x"),
                CellFormat(mnt.mnt.mnt_sb.value_(), "^x"),
                escape_ascii_string(mount_fstype(mnt), escape_backslash=True),
                escape_ascii_string(mount_src(mnt), escape_backslash=True),
                escape_ascii_string(mount_dst(mnt), escape_backslash=True),
            )
        )
    print_table(rows)
